@@ -1,5 +1,6 @@
-#!/usr/bin/python2
+#!/usr/bin/python3
 
+import collections
 import os
 import glob
 import shutil
@@ -14,27 +15,49 @@ try:
     ENVIRON = os.environ['XDG_CURRENT_DESKTOP']
 except:
     ENVIRON = ""
+
 D_GROUP = "Desktop Entry"
 DEFAULT_ICON = "system-run"
-AUTOSTART_APPS = []
+AUTOSTART_APPS = collections.OrderedDict()
+
+KEYFILE_FLAGS = GLib.KeyFileFlags.KEEP_COMMENTS and GLib.KeyFileFlags.KEEP_TRANSLATIONS
 
 def list_header_func(row, before, user_data):
     if before and not row.get_header():
         row.set_header(Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL))
 
+
+def get_appname(filename):
+    """ Get the desktop basename without the .desktop """
+    basename = os.path.basename(filename)
+    if basename.endswith('.desktop'):
+        basename = basename[:-len('.desktop')]
+    return basename
+
+
+def get_blacklisted_apps():
+    source = Gio.SettingsSchemaSource.get_default()
+    schema = source.lookup('org.cinnamon.SessionManager', True)
+    if not schema or not schema.has_key('autostart-blacklist'):  # noqa
+        return {}
+
+    cs_settings = Gio.Settings('org.cinnamon.SessionManager')
+    return set(cs_settings['autostart-blacklist'])
+
+
 class Module:
     name = "startup"
     comment = _("Manage your startup applications")
     category = "prefs"
 
     def __init__(self, content_box):
-        keywords = _("startup, programs, boot, init, session")
+        keywords = _("startup, programs, boot, init, session, autostart, apps")
         sidePage = SidePage(_("Startup Applications"), "cs-startup-programs", keywords, content_box, module=self)
         self.sidePage = sidePage
 
     def on_module_selected(self):
         if not self.loaded:
-            print "Loading Startup Applications module"
+            print("Loading Startup Applications module")
 
             page = SettingsPage()
             page.expand = True
@@ -47,7 +70,7 @@ class Module:
 
             self.gather_apps()
 
-            for app in AUTOSTART_APPS:
+            for app in AUTOSTART_APPS.values():
                 if app.key_file_loaded and app.shown and not app.no_display and not app.hidden:
                     row = AutostartRow(app)
                     settings.add_row(row)
@@ -58,28 +81,32 @@ class Module:
             try:
                 os.makedirs(user_autostart_dir)
             except:
-                print "Could not create autostart dir: %s" % user_autostart_dir
+                print("Could not create autostart dir: %s" % user_autostart_dir)
 
     def gather_apps(self):
         system_files = []
 
+        blacklisted_apps = get_blacklisted_apps()
+
         user_files = glob.glob(os.path.join(GLib.get_user_config_dir(), "autostart", "*.desktop"))
         for app in user_files:
-            AUTOSTART_APPS.append(AutostartApp(app, user_position=os.path.dirname(app)))
+            key = get_appname(app)
+            if key in blacklisted_apps:
+                continue
+            AUTOSTART_APPS[key] = AutostartApp(app, user_position=os.path.dirname(app))
 
         for d in GLib.get_system_config_dirs():
             system_files.extend(glob.glob(os.path.join(d, "autostart", "*.desktop")))
 
         for sys_app in system_files:
-            found = False
-            for app in AUTOSTART_APPS:
-                if os.path.basename(sys_app) == os.path.basename(app.app):
-                    app.system_position = os.path.dirname(sys_app)
-                    found = True
-                    break
-
-            if not found:
-                AUTOSTART_APPS.append(AutostartApp(sys_app, system_position=os.path.dirname(sys_app)))
+            key = get_appname(sys_app)
+            if key in blacklisted_apps:
+                continue
+            if key in AUTOSTART_APPS:
+                AUTOSTART_APPS[key].system_position = os.path.dirname(sys_app)
+            else:
+                AUTOSTART_APPS[key] = AutostartApp(sys_app,
+                                                   system_position=os.path.dirname(sys_app))
 
 class AutostartApp():
     def __init__(self, app, user_position=None, system_position=None):
@@ -97,9 +124,9 @@ class AutostartApp():
 
     def load(self):
         try:
-            self.key_file.load_from_file(self.app, GLib.KeyFileFlags.KEEP_COMMENTS and GLib.KeyFileFlags.KEEP_TRANSLATIONS)
-        except GLib.GError:
-            print "Failed to load %s" % self.app
+            self.key_file.load_from_file(self.app, KEYFILE_FLAGS)
+        except GLib.GError as e:
+            print("Failed to load %s" % self.app, e)
             return
 
         self.key_file_loaded = True
@@ -150,7 +177,7 @@ class AutostartApp():
         if only_show_in:
             found = False
             for i in only_show_in:
-                if i == ENVIRON:
+                if i in ('GNOME', 'X-Cinnamon'):
                     found = True
                     break
             if not found:
@@ -181,20 +208,20 @@ class AutostartApp():
             self.app = os.path.join(self.system_position, self.basename)
             os.remove(old_app)
             self.key_file = GLib.KeyFile.new()
-            if self.key_file.load_from_file(self.app, GLib.KeyFileFlags.KEEP_COMMENTS and GLib.KeyFileFlags.KEEP_TRANSLATIONS):
+            if self.key_file.load_from_file(self.app, KEYFILE_FLAGS):
                 self.load()
             self.save_done_success()
             return False
 
-        if self.system_position:
-            use_path = os.path.join(self.system_position, self.basename)
-        else:
-            use_path = self.path
-
-        key_file = GLib.KeyFile.new()
-
         try:
-            key_file.load_from_file(use_path, GLib.KeyFileFlags.KEEP_COMMENTS and GLib.KeyFileFlags.KEEP_TRANSLATIONS)
+            key_file = GLib.KeyFile.new()
+
+            if self.user_position == None:
+                self.user_position = os.path.join(GLib.get_user_config_dir(), "autostart")
+                self.path = os.path.join(self.user_position, self.basename)
+                key_file.load_from_file(os.path.join(self.system_position, self.basename), KEYFILE_FLAGS)
+            else:
+                key_file.load_from_file(self.path, KEYFILE_FLAGS)
         except:
             key_file.set_string(D_GROUP, GLib.KEY_FILE_DESKTOP_KEY_TYPE, "Application")
             key_file.set_string(D_GROUP, GLib.KEY_FILE_DESKTOP_KEY_EXEC, "/bin/false")
@@ -235,13 +262,6 @@ class AutostartApp():
 
         return False
 
-    def queue_save(self):
-        if self.user_position == None:
-            self.user_position = os.path.join(GLib.get_user_config_dir(), "autostart")
-            self.path = os.path.join(self.user_position, self.basename)
-
-        self.save()
-
     def update(self, info):
         changed = False
         if info["name"] != self.name:
@@ -268,12 +288,12 @@ class AutostartApp():
             changed = True
 
         if changed:
-            self.queue_save()
+            self.save()
 
     def set_enabled(self, enabled):
         self.enabled = enabled
         self.save_mask.add_item("enabled")
-        self.queue_save()
+        self.save()
 
     def remove(self):
         if not self.system_position and self.user_position:
@@ -283,13 +303,13 @@ class AutostartApp():
             self.save_mask.add_item("hidden")
             self.enabled = False
             self.save_mask.add_item("enabled")
-            self.queue_save()
+            self.save()
 
     def update_description(self):
         if self.name == "":
             self.name = _("No name")
         if self.comment == "":
-            self.comment == _("No description")
+            self.comment = _("No description")
 
     def user_equals_system(self):
         if not self.system_position:
@@ -302,7 +322,7 @@ class AutostartApp():
 
         if (self.get_boolean(key_file, GLib.KEY_FILE_DESKTOP_KEY_HIDDEN, False) != self.hidden or
             self.get_boolean(key_file, "X-GNOME-Autostart-enabled", True) != self.enabled or
-            self.get_shown(key_file) != self.shown):
+                self.get_shown(key_file) != self.shown):
             return False
 
         if self.get_boolean(key_file, GLib.KEY_FILE_DESKTOP_KEY_NO_DISPLAY, False) != self.no_display:
@@ -373,16 +393,6 @@ class AutostartBox(Gtk.Box):
         toolbar.add(title_holder)
         main_box.add(toolbar)
 
-        toolbar_separator = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
-        main_box.add(toolbar_separator)
-        separator_context = toolbar_separator.get_style_context()
-        frame_color = frame_style.get_border_color(Gtk.StateFlags.NORMAL).to_string()
-        css_provider = Gtk.CssProvider()
-        css_provider.load_from_data(".separator { -GtkWidget-wide-separators: 0; \
-                                                   color: %s;                    \
-                                                }" % frame_color)
-        separator_context.add_provider(css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
-
         scw = Gtk.ScrolledWindow()
         scw.expand = True
         scw.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
@@ -415,24 +425,28 @@ class AutostartBox(Gtk.Box):
         box.set_halign(Gtk.Align.CENTER)
         button_holder.add(box)
 
-        self.add_button = Gtk.Button.new_with_label(_("Add"))
+        self.add_button = Gtk.Button.new_from_icon_name("list-add-symbolic", Gtk.IconSize.BUTTON)
+        self.add_button.set_tooltip_text(_("Add"))
         self.add_button.connect("clicked", self.on_add_button_clicked)
         button_group.add_widget(self.add_button)
         box.add(self.add_button)
 
-        self.edit_button = Gtk.Button.new_with_label(_("Edit"))
+        self.edit_button = Gtk.Button.new_from_icon_name("document-edit-symbolic", Gtk.IconSize.BUTTON)
+        self.edit_button.set_tooltip_text(_("Edit"))
         self.edit_button.connect("clicked", self.on_edit_button_clicked)
         button_group.add_widget(self.edit_button)
         self.edit_button.set_sensitive(False)
         box.add(self.edit_button)
 
-        self.remove_button = Gtk.Button.new_with_label(_("Remove"))
+        self.remove_button = Gtk.Button.new_from_icon_name("list-remove-symbolic", Gtk.IconSize.BUTTON)
+        self.remove_button.set_tooltip_text(_("Remove"))
         self.remove_button.connect("clicked", self.on_remove_button_clicked)
         button_group.add_widget(self.remove_button)
         self.remove_button.set_sensitive(False)
         box.add(self.remove_button)
 
-        self.run_button = Gtk.Button.new_with_label(_("Run now"))
+        self.run_button = Gtk.Button.new_from_icon_name("system-run-symbolic", Gtk.IconSize.BUTTON)
+        self.run_button.set_tooltip_text(_("Run now"))
         self.run_button.connect("clicked", self.on_run_button_clicked)
         button_group.add_widget(self.run_button)
         self.run_button.set_sensitive(False)
@@ -529,9 +543,11 @@ class AutostartBox(Gtk.Box):
                 return
 
             app = AutostartApp(filename, user_position=os.path.dirname(filename))
+            key = get_appname(filename)
+            AUTOSTART_APPS[key] = app
 
             app.basename = os.path.basename(app.app)
-            app.dir = os.path.basename(app.app)
+            app.dir = os.path.dirname(app.app)
             app.hidden = False
             app.no_display = False
             app.enabled = True
@@ -547,7 +563,7 @@ class AutostartBox(Gtk.Box):
 
             app.save_mask.add_item("all")
 
-            app.queue_save()
+            app.save()
 
             row = AutostartRow(app)
             self.add_row(row)
@@ -570,15 +586,17 @@ class AutostartBox(Gtk.Box):
             try:
                 shutil.copyfile(desktop_file_dir, user_desktop_file)
             except IOError:
-                print "Failed to copy desktop file %s" % desktop_file_name
+                print("Failed to copy desktop file %s" % desktop_file_name)
 
             app = AutostartApp(user_desktop_file, user_position=os.path.dirname(user_desktop_file))
+            key = get_appname(user_desktop_file)
+            AUTOSTART_APPS[key] = app
 
             app.enabled = True
 
             app.save_mask.add_item("all")
 
-            app.queue_save()
+            app.save()
 
             row = AutostartRow(app)
             self.add_row(row)
@@ -610,11 +628,8 @@ class AutostartBox(Gtk.Box):
         return filename
 
     def find_app_with_basename(self, basename):
-        for app in AUTOSTART_APPS:
-            if basename == app.basename:
-                return app
-
-        return None
+        key = get_appname(basename)
+        return AUTOSTART_APPS.get(key)
 
     def popup_menu_below_button (self, *args):
         # the introspection for GtkMenuPositionFunc seems to change with each Gtk version,
@@ -649,15 +664,14 @@ class AutostartRow(Gtk.ListBoxRow):
         if self.app.icon:
             try:
                 if GLib.path_is_absolute(self.app.icon):
-                    shown_icon = GdkPixbuf.Pixbuf.new_from_file_at_scale(self.app.icon, 24, 24, True)
-                    img = Gtk.Image.new_from_pixbuf(shown_icon)
-                else:
-                    pixbuf = icon_theme.load_icon(self.app.icon, 24, Gtk.IconLookupFlags.FORCE_SIZE)
+                    pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_scale(self.app.icon, 24, 24, True)
                     img = Gtk.Image.new_from_pixbuf(pixbuf)
+                else:
+                    img = Gtk.Image.new_from_icon_name(self.app.icon, Gtk.IconSize.LARGE_TOOLBAR)
             except:
-                img = Gtk.Image.new_from_gicon(Gio.ThemedIcon.new(DEFAULT_ICON), Gtk.IconSize.LARGE_TOOLBAR)
+                img = Gtk.Image.new_from_icon_name(DEFAULT_ICON, Gtk.IconSize.LARGE_TOOLBAR)
         else:
-            img = Gtk.Image.new_from_gicon(Gio.ThemedIcon.new(DEFAULT_ICON), Gtk.IconSize.LARGE_TOOLBAR)
+            img = Gtk.Image.new_from_icon_name(DEFAULT_ICON, Gtk.IconSize.LARGE_TOOLBAR)
         grid.attach(img, 0, 0, 1, 1)
 
         self.desc_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
@@ -681,16 +695,21 @@ class AutostartRow(Gtk.ListBoxRow):
         self.delay_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
         self.delay_box.props.hexpand = False
         self.delay_box.set_margin_right(15)
+        delay_time_markup = GLib.markup_escape_text(self.app.delay)
+
         label = Gtk.Label(_("Delay"))
         self.delay_box.pack_start(label, False, False, 0)
         self.delay_time_label = Gtk.Label()
-        delay_time_markup = GLib.markup_escape_text(self.app.delay)
-        self.delay_time_label.set_markup(delay_time_markup)
+        self.delay_time_label.set_markup(_("%s s") % delay_time_markup)
         self.delay_time_label.get_style_context().add_class("dim-label")
         self.delay_box.pack_start(self.delay_time_label, False, False, 0)
         grid.attach_next_to(self.delay_box, self.desc_box, Gtk.PositionType.RIGHT, 1, 1)
         self.delay_box.set_sensitive(app.enabled)
 
+        self.delay_box.show_all()
+        self.delay_box.set_no_show_all(True)
+        self.delay_box.set_visible(delay_time_markup != "0")
+
         switch_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
         switch_box.set_margin_top(5)
         switch_box.set_margin_bottom(5)
@@ -709,7 +728,8 @@ class AutostartRow(Gtk.ListBoxRow):
 
         self.name_label.set_markup("<b>{}</b>".format(name_markup))
         self.comment_label.set_markup("<small>{}</small>".format(comment_markup))
-        self.delay_time_label.set_markup(delay_time_markup)
+        self.delay_time_label.set_markup(_("%s s") % delay_time_markup)
+        self.delay_box.set_visible(delay_time_markup != "0")
 
     def on_switch_activated(self, switch, gparam):
         active = switch.get_active()
